# -*- coding: utf-8 -*-

__author__    = "renaud blanch <http://iihm.imag.fr/blanch/>"
__copyright__ = "Copyright © 2010-2011 - Renaud Blanch"
__licence__   = "GPLv3 <http://www.gnu.org/licenses/gpl.html>"


"""
Experiment flow implementation.

see <http://iihm.imag.fr/blanch/projects/distractors/>
"""


# imports ####################################################################

import sys

from time import time

from OpenGL.GLUT import *
from OpenGL.GL import *


# opengl initialisation ######################################################

glutInit(sys.argv)
glutInitDisplayMode(GLUT_RGBA|GLUT_DOUBLE)
glutCreateWindow("expe")
glutSetCursor(GLUT_CURSOR_NONE)


# experiment #################################################################

from _input import *

from constants import *
from graphics import *
from task import Task
from techniques import Flat


# state ######################################################################

_start_time_s = time()
def ms(s):
	return int((s-_start_time_s)*1000.)

ANIMATION_T = 20 # ms

class Experiment(object):
	def __init__(self, trials):
		self.trials = trials
		self.O = 0, 0
		self.message = None
		self.time = time()
		
		self.passed_count = 0
		self.trial_count = 0
		
		self.init_glut()
		self.init_state()
		self.init_input()

	def run(self):
		glutPositionWindow(2*SCREEN_WIDTH, 0) # force window on second screen
		glutFullScreen()
		glutMainLoop()


	# glut
	
	def init_glut(self):
		glutReshapeFunc(self.reshape)
		glutDisplayFunc(self.display)
		
	def reshape(self, W, H):
		glViewport(0, 0, W, H)
		glMatrixMode(GL_PROJECTION)
		glLoadIdentity()
		glOrtho(0, W, H, 0, -1, 1)
		glMatrixMode(GL_MODELVIEW)
		glLoadIdentity()
		glTranslate(0, H/2, 0)
		glTranslate(X_OFFSET, 0, 0)

	def display(self):
		glClear(GL_COLOR_BUFFER_BIT)
		self.state.draw()
		glutSwapBuffers()

	def timer(self, registration_time_ms):
		if ms(self.time) > registration_time_ms:
			return		
		self.state.idle(time() - self.time)
		glutTimerFunc(ANIMATION_T, self.timer, registration_time_ms)
		glutPostRedisplay()


	# state
	
	def init_state(self):
		self.Technique, task_args, self.training = self.trials.next()
		self.task = Task(*task_args)
		self.state = Waiting(self.start, self.task, X=500)

	def start(self, X, Y):
		if self.task.dimension == 1:
			self.O = X, 0, 0
		elif self.task.dimension == 2:
			self.O = X, Y, 0
		glPushMatrix()
		glTranslate(*self.O)
		self.time = time()
		self.state = Running(self.stop, self.task, self.Technique, 
		                     self.O, self.time, self.training)
		
	def stop(self, passed, X, Y):
		self.trial_count += 1
		if passed:
			self.passed_count += 1
		
		glPopMatrix()
		dX, dY, _ = self.O

		trial = self.trials.next()
		if isinstance(trial, unicode):
			passed = 100. * self.passed_count / self.trial_count
			error = 100. - passed
			message = u"Votre taux d'erreur pour le bloc précédent est de : %0.1f%%.\n" % error
			if error < 4.:
				message += u"Vous pouvez aller plus vite."
			elif error < 5.:
				message += u"Continuez comme ça."
			else:
				message += u"Vous pouvez faire mieux."
			self.trial_count, self.passed_count = 0, 0
			
			while isinstance(trial, unicode):
				message = u"\n".join([message, trial])
				try:
					trial = self.trials.next()
				except StopIteration:
					sys.exit(0)
		else:
			message = None
		
		self.Technique, task_args, self.training = trial
		self.task = Task(*task_args)

		self.state = Waiting(self.start, self.task, X+dX, Y+dY, message)
		

	# input
	
	def init_input(self):
		set_motion_cb(self.motion)
		set_button_cb(self.button)

	def motion(self, dx, dy):
		self.time = time()
		self.state.motion(dx, dy, self.time)
		glutTimerFunc(ANIMATION_T, self.timer, ms(self.time))
		glutPostRedisplay()
	
	def button(self, down):
		if down:
			self.state.press(time())
		glutPostRedisplay()
		

# experiment states ##########################################################

WAITING_T = 1.  # s
WAITING_D = 10. # px
HOME = (WAITING_D, 0., 0.)

class Waiting(object):
	def __init__(self, start, task, X, Y=0., message=None):
		self.start = start
		self.dimension = task.dimension
		self.technique = Flat([HOME], ([0.], {0.:[0]}),
		                      task.distance, task.dimension)
		self.technique.X, self.technique.Y = X, Y
		self.message = message
		
	def draw(self):
		self.technique.feedback()
		draw_target(self.dimension, color=HOME_COLOR, *HOME)
		self.technique.cursor()
		if self.message:
			draw_text(self.message)
		
	def motion(self, dx, dy, t):
		self.technique.motion(dx, dy)

	def press(self, t):
		pass
		
	def idle(self, duration):
		if duration < WAITING_T:
			return
		X, Y = self.technique.pick()
		if self.technique.distance(X, Y, *HOME) > 0.:
			return
		self.start(X, Y)
		
	
class Running(object):
	def __init__(self, stop, task, Technique, origin, t, training):
		self.stop = stop
		self.task = task
		self.technique = Technique(self.task.distractors,
		                           self.task.distances,
		                           self.task.distance,
		                           self.task.dimension,
		                           origin)
		self.x, self.y = 0, 0
		self.start_t = self.t = t
		print "# %s %s %s" % (self.task, self.technique, training)

	def draw(self):
		self.technique.feedback()
		draw_task(self.task)
		self.technique.cursor()

	def motion(self, dx, dy, t):
		self.x += dx
		self.y += dy
		self.technique.motion(dx, dy, t-self.t)
		self.t = t
		print t-self.start_t, self.x, self.y, self.technique.X, self.technique.Y
	
	def press(self, t):
		passed = self.task.inside(*self.technique.pick())
		print t-self.start_t, passed		
		self.stop(passed, self.technique.X, self.technique.Y)
					
	def idle(self, duration):
		self.technique.idle(duration)


# main #######################################################################

if __name__ == "__main__":
	from techniques import (Flat, SemanticPointing, BubbleCursor, 
	                        DynaSpot, RakeCursor)

	def trials():
		for Technique in [BubbleCursor]:
			for d in range(4):
				R = .6/(2**d)
				for ID in [3, 4, 5]:
					for A in [511, 1023]:
						yield Technique, (ID, R, A), False
			yield u"Nouvelle technique."

	experiment = Experiment(trials())
	experiment.run()

