# -*- coding: utf-8 -*-

__author__    = "renaud blanch <http://iihm.imag.fr/blanch/>"
__copyright__ = "Copyright © 2010-2011 - Renaud Blanch"
__licence__   = "GPLv3 <http://www.gnu.org/licenses/gpl.html>"


"""
Graphical backend.

see <http://iihm.imag.fr/blanch/projects/distractors/>
"""


# imports ####################################################################

import os

from math import ceil, floor

from PIL import Image

from OpenGL.GL import *
from OpenGL.GLUT import *

from constants import *


# shader utils ###############################################################

class ShaderCompileError(RuntimeError):
	"""shader compile error."""
	
class ProgramLinkError(RuntimeError):
	"""program link error."""


def create_shader(shader_source, shader_type):
	"""creates a shader from its source & type."""
	shader = glCreateShader(shader_type)
	glShaderSource(shader, shader_source)
	glCompileShader(shader)
	if glGetShaderiv(shader, GL_COMPILE_STATUS) != GL_TRUE:
		raise ShaderCompileError(glGetShaderInfoLog(shader))
	return shader

def create_program(*shaders):
	"""creates a program from its vertex & fragment shader sources."""
	program = glCreateProgram()
	for shader in shaders:
		glAttachShader(program, shader)
	glLinkProgram(program)
	if glGetProgramiv(program, GL_LINK_STATUS) != GL_TRUE:
		raise ProgramLinkError(glGetProgramInfoLog(program))
	return program


# globals ####################################################################

glEnable(GL_BLEND)
glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)

glClearColor(*BACKGROUND_COLOR)


# circle #####################################################################

target_2d_program = create_program(
	create_shader("""\
	void main() {
		vec2 p = gl_TexCoord[0].st;
		float v = length(dFdx(p)+dFdy(p));
		if(length(p) > 1.+v) discard;
		float alpha = clamp((1.-length(p))/v, 0., 1.);
		gl_FragColor = vec4(gl_Color.rgb, alpha*gl_Color.a);
	}
	""", GL_FRAGMENT_SHADER)
)

def target_1d(R, X, color=FEEDBACK_COLOR):
	half_height = SCREEN_HEIGHT/2.
	r, g, b, a = color

	if R <= .5:
		glColor4f(r, g, b, a*2*R)
		glBegin(GL_LINES)
		glVertex(X-.5, -half_height)
		glVertex(X-.5, half_height)
		glEnd()

	else:
		glColor(r, g, b, a)

		X0 = X - R
		X1 = X0 + 2*R
		X0i, X1i = ceil(X0), floor(X1)
		glRectf(X0i, -half_height, X1i, half_height)

		a0, a1 = X0i - X0, X1 - X1i
	
		X0i -= .5
		X1i += .5
		glBegin(GL_LINES)
		glColor4f(r, g, b, a*a0)
		glVertex2f(X0i, -half_height)
		glVertex2f(X0i,  half_height)

		glColor4f(r, g, b, a*a1)
		glVertex2f(X1i, -half_height)
		glVertex2f(X1i,  half_height)
		glEnd()


def target_2d(R, X, Y=0., color=FEEDBACK_COLOR):
	if R < .5:
		return
	if abs(Y)-2*R > SCREEN_HEIGHT/2.:
		return
	if X-2*R > SCREEN_WIDTH-X_OFFSET:
		return
	glColor(*color)
	glBegin(GL_TRIANGLE_STRIP)
	for s in [-1., +1.]:
		for t in [-1., +1.]:
			glTexCoord(s, t)
			glVertex(X+s*R, Y-t*R)
	glEnd()

def draw_target(dimension, R, X, Y=0., color=FEEDBACK_COLOR):
	if dimension == 1:
		target_1d(R, X, color)
	elif dimension == 2:
		glUseProgram(target_2d_program)
		target_2d(R, X, Y, color)
		glUseProgram(0)


# cursor #####################################################################

PIL_FORMATS = {
	"RGB":  GL_RGB,
	"RGBA": GL_RGBA,
	"LA":   GL_LUMINANCE_ALPHA,
}

cursor = os.path.join(os.path.dirname(__file__), "cursor.png")
cursor = Image.open(cursor)
cursor_X, cursor_Y = 4, 4
cursor_W, cursor_H = cursor.size
cursor_format = PIL_FORMATS[cursor.mode]
cursor_pixels = cursor.tostring()

glPixelZoom(1., -1.)
glPixelStorei(GL_UNPACK_ALIGNMENT, 1)

def draw_cursor(dimension, X, Y=0., alpha=1.):
	if dimension == 1:
		r, g, b, a = CURSOR_COLOR
		glColor(r, g, b, a*alpha)
		glRectf(X-.5, -SCREEN_HEIGHT, X+.5, SCREEN_HEIGHT)
	elif dimension == 2:
		glPixelTransferf(GL_ALPHA_SCALE, alpha)
		glRasterPos(0, 0)
		glBitmap(0, 0, 0, 0, X-cursor_X, -(Y-cursor_Y), None)
		glDrawPixels(cursor_W, cursor_H, 
		             cursor_format, GL_UNSIGNED_BYTE,
		             cursor_pixels)
		glPixelTransferf(GL_ALPHA_SCALE, 1.)


# text #######################################################################

def draw_text(s, X=0, Y=50):
	glColor(*TEXT_COLOR)
	for i, line in enumerate(s.split('\n')):
		glRasterPos(X, Y+20*i)
		for c in line:
			glutBitmapCharacter(GLUT_BITMAP_8_BY_13, ord(c))


# task #######################################################################

_task = None
_task_id = glGenLists(1)

def draw_task(task):
	global _task
	if task != _task:
		_task = task
		target = {
			1: target_1d,
			2: target_2d,
		}[task.dimension]

		glNewList(_task_id, GL_COMPILE_AND_EXECUTE)
		if target is target_2d:
			glUseProgram(target_2d_program)
		for distractor in task.distractors:
			target(color=DISTRACTOR_COLOR, *distractor)
		target(color=TARGET_COLOR, *task.target)
		glUseProgram(0)
		glEndList()
	else:	
		glCallList(_task_id)
