# -*- coding: utf-8 -*-

__author__    = "renaud blanch <http://iihm.imag.fr/blanch/>"
__copyright__ = "Copyright © 2010-2011 - Renaud Blanch"
__licence__   = "GPLv3 <http://www.gnu.org/licenses/gpl.html>"


"""
Task definition

see <http://iihm.imag.fr/blanch/projects/distractors/>
"""


# imports ####################################################################

from math import asin, sqrt, pi, tan, sin, cos, floor, ceil, log, hypot

from constants import *


# target size as functions of ID and amplitude ###############################

def W_Fitts(ID, A):
	"""target width as function of ID and task amplitude [Fitts]."""
	return A / (2**(ID-1))

def W_MacKenzie(ID, A):
	"""target width as function of ID and task amplitude [MacKenzie]."""
	return A / (2**ID - 1)


# distances ##################################################################

def distance_1d(Xc, Yc=None, R=0., X=0., Y=None):
	return abs(X-Xc) - R

def distance_2d(Xc, Yc=0., R=0., X=0., Y=0.):
	return hypot(X-Xc, Y-Yc) - R
		

# distractors layout #########################################################

SHRINK_T = sin(pi/3) * pi/4 # pi/4: circle density; sin(pi/3): hex packing

def layout_1d_lin(R, A, W):
	dA = W/R
	i_min = int(floor(-(A+X_OFFSET)/dA))
	i_max = i_min + int(ceil(SCREEN_WIDTH/dA)) + 1
	for i in range(i_min, i_max+1):
		yield W/2., A+i*dA
	
def layout_2d_lin(R, A, W):
	dA = W/2*sqrt(pi/(sin(pi/3)*R))
	i_min = int(floor(-(A+X_OFFSET)/dA))
	i_max = i_min + int(ceil(SCREEN_WIDTH/dA)) + 1
	dY = dA*sin(pi/3)
	n = int(floor(SCREEN_HEIGHT/2./dY)) + 1
	for j in range(-n, n+1):
		if j%2 == 0:
			for i in range(i_min, i_max+1):
				yield W/2., A+i*dA, j*dY
		else:
			for i in range(i_min, i_max):
				yield W/2., A+(i+.5)*dA, j*dY


def layout_1d_log(R, A, W):
	r = W/A/2.
	k = (R + r)/(R - r)
	i_max = int(floor(log(MAX_A/A) / log(k)))
	i_min = int( ceil(log(MIN_A/A) / log(k)))
	for i in range(i_min, i_max+1):
		ki = k**i
		a, w = ki*A, ki*W
		yield w/2., a
	
def layout_2d_log(R, A, W):
	# splitting 2D density across tangential and radial 1D densities
	try:
		r_t = sqrt(R/SHRINK_T)
		r_r = R/r_t
	except:
		r_r, r_t = 0., 0.

	# computing scaling factor and circular slice range
	alpha = asin(W/A/2)
	try:
		k = sqrt(pi/(r_r/tan(alpha) + alpha - pi/2) + 1)
	except ValueError:
		raise StopIteration
	i_max = int(floor(log(MAX_A/A) / log(k)))
	i_min = int( ceil(log(MIN_A/A) / log(k)))
	
	# computing sectors number
	n = int(ceil(r_t*pi/alpha/4))
	
	for i in range(i_min, i_max+1):
		ki0 = k**i
		ki1 = k**(i+.5)
		a0, w0 = ki0*A, ki0*W
		a1, w1 = ki1*A, ki1*W
		for j in range(-n, n+1):
			angle = 2*alpha/r_t*j
			if (j%2) == 0:
				yield w0/2., a0*cos(angle), a0*sin(angle)
			else:
				yield w1/2., a1*cos(angle), a1*sin(angle)

		
# pointing task ##############################################################

class Task(object):
	"""pointing task."""
	def __init__(self, ID, R, A, dimension=2, linear=False, W=W_MacKenzie):
		self.ID = float(ID)
		self.R = float(R)
		self.A = float(A)

		self.W = W(self.ID, self.A)
		self.target = (self.W/2., self.A)
		
		self.dimension = dimension
		self.linear    = linear
		layout = {
			(1, True):  layout_1d_lin,
			(2, True):  layout_2d_lin,
			(1, False): layout_1d_log,
			(2, False): layout_2d_log,
		}[(self.dimension, self.linear)]
		
		self.distractors = list(layout(self.R, self.A, self.W))
		if self.target not in self.distractors:
			self.distractors.append(self.target)
		
		self.distance = {
			1: distance_1d,
			2: distance_2d,
		}[self.dimension]

		distances = {}
		for i, D in enumerate(self.distractors):
			H = self.distance(*D[1:])
			H = float(str(H))
			indicies = distances.get(H, [])
			indicies.append(i)
			distances[H] = indicies
		sorted_keys = list(distances)
		sorted_keys.sort()
		self.distances = (sorted_keys, distances)
		
		
	def __repr__(self):
		return "%(ID)s %(R)s %(A)s %(dimension)sD %(linear)s" % self.__dict__
	
	def inside(self, X, Y=0.):
		return self.distance(X, Y, self.W/2., self.A, 0.) <= 0.
	
	def _key(self):
		return self.ID, self.R, self.A, self.W, self.dimension, self.linear
	
	def __ne__(self, other):
		try:
			return self._key() != other._key()
		except:
			return NotImplemented
