# -*- coding: utf-8 -*-

__author__    = "renaud blanch <http://iihm.imag.fr/blanch/>"
__copyright__ = "Copyright © 2010-2011 - Renaud Blanch"
__licence__   = "GPLv3 <http://www.gnu.org/licenses/gpl.html>"


"""
Some pointing techniques implementation:
- Flat is normal pointing without acceleration
- SemanticPointing is the semantic pointing technique
- BubbleCursor is the bubble cursor technique
- DynaSpot is the BynaSpot technique
- RakeCursor is the rake cursor technique
(see the paper for the proper references)

see <http://iihm.imag.fr/blanch/projects/distractors/>
"""


# imports ####################################################################

from math import hypot, exp, sin, pi, hypot, floor, ceil

from constants import *
from graphics import draw_target, draw_cursor

	
# flat technique #############################################################

def indicies(distance, X, Y, sorted_keys):
	Hc = distance(X, Y)
	i_min, i_max = 0, len(sorted_keys)
	for i, k in enumerate(sorted_keys):
		if k < Hc:
			i_min = max(i-1, i_min)
		elif k > Hc:
			i_max = min(i+2, i_max)
			break
	return slice(i_min, i_max)


class Flat(object):
	DEFAULT_GAIN = 2. # dX/dx
	DEFAULT_MARGIN = 4.

	def __init__(self, distractors, distances, distance, dimension,
	             origin=(0, 0, 0), margin=None, X=0., Y=0.):
		self.distractors = distractors
		self.distances = distances
		self.distance = distance
		self.dimension = dimension
		self.origin = origin
		self.margin = margin or self.DEFAULT_MARGIN
		self.X, self.Y = X, Y
		self.gain = self.DEFAULT_GAIN
		self.motion()

	
	# proximity tests
	
	def closest(self):
		X, Y = self.X, self.Y
		distractors = self.distractors
		sorted_keys, distances = self.distances
		distance = self.distance

		D0 = distractors[0]
		L0 = distance(X, Y, *D0)
		for k in sorted_keys[indicies(distance, X, Y, sorted_keys)]:
			for i in distances[k]:
				D = distractors[i]
				L = distance(X, Y, *D)
				if L < L0:
					D0, L0 = D, L
		return D0, L0
	
	def closest_2(self):
		X, Y = self.X, self.Y
		distractors = self.distractors
		sorted_keys, distances = self.distances
		distance = self.distance

		D0 = distractors[0]
		L0 = distance(X, Y, *D0)
		D1 = distractors[1]
		L1 = distance(X, Y, *D1)
		for k in sorted_keys[indicies(distance, X, Y, sorted_keys)]:
			for i in distances[k]:
				D = distractors[i]
				L = distance(X, Y, *D)
				if L < L0:
					D1, L1 = D0, L0
					D0, L0 = D,  L
				elif L < L1:
					D1, L1 = D, L
		return (D0, L0), (D1, L1)


	# technique implementation
	
	def motion(self, dx=0, dy=0, dt=1.):
		dX, dY = self.dP(dx, dy)
		self.X += dX
		self.Y += dY
	
	def dP(self, dx, dy):
		return dx*PX_PER_PT*self.gain, dy*PX_PER_PT*self.gain
		
	def idle(self, duration):	
		pass

	def pick(self):
		return self.X, self.Y

	def feedback(self):
		D, L = self.closest()
		if L <= 0:
			draw_target(self.dimension, 
			            D[0]+self.margin, *D[1:])
	
	def cursor(self):
		draw_cursor(self.dimension, self.X, self.Y)

	def __repr__(self):
		return self.__class__.__name__

	
# semantic pointing technique ################################################

class SemanticPointing(Flat):
	DEFAULT_SCALE = 4.
	def __init__(self, distractors, distances, distance, dimension, origin,
	                   scale=None, **kwargs):
		self.scale = scale or self.DEFAULT_SCALE
		super(SemanticPointing, self).__init__(distractors, distances, distance, 
		                                       dimension, origin, **kwargs)
		self.gain *= self.scale/2.
		
	def motion(self, dx=0, dy=0, dt=1.):
		dX, dY = self.dP(dx, dy)

		_, L = self.closest()
		margin = self.margin*2.
		if L > margin:
			scale = 1.
		elif L > -margin:
			scale = 1.+(self.scale-1.)*(margin-L)/(2.*margin)
		else:
			scale = self.scale

		self.X += dX/scale
		self.Y += dY/scale


# bubble cursor technique ####################################################

class BubbleCursor(Flat):
	def pick(self):
		D, _ = self.closest()
		return D[1:]

	def feedback(self):
		try:
			(D0, L0), (D1, L1) = self.closest_2()
		except IndexError:
			D0, L0 = self.closest()
			R0 = D0[0]
			R = L0 + 2.*R0
		else:
			R0 = D0[0]
			R = min(L0+2.*R0, L1)
		draw_target(self.dimension, R0+self.margin, *D0[1:])
		draw_target(self.dimension, R, self.X, self.Y)


# dynaspot technique #########################################################

class DynaSpot(Flat):
	MAX_SPOTWIDTH = 32.  # px
	MIN_SPEED     = 100. # px/s
	
	def __init__(self, distractors, distances, distance, 
	                   dimension, origin, **kwargs):
		self.spot_width = 1.
		super(DynaSpot, self).__init__(distractors, distances, distance,
		                               dimension, origin, **kwargs)
	
	def motion(self, dx=0, dy=0, dt=1.):
		dX, dY = self.dP(dx, dy)
		
		if hypot(dX, dY)/dt > self.MIN_SPEED:
			self.spot_width = min(self.spot_width*1.2, self.MAX_SPOTWIDTH)

		self.X += dX
		self.Y += dY

	def idle(self, duration):
		if duration < 0.120:
			max_spot_width = self.spot_width
		elif duration < 0.300:
			max_spot_width = (self.MAX_SPOTWIDTH-1.) * (1. - (exp((duration-0.120)/0.180)-1.)/(exp(1.)-1.)) + 1.
		else:
			max_spot_width = 1.
		self.spot_width = min(self.spot_width, max_spot_width)
	
	def pick(self):
		D, L = self.closest()
		if L <= self.spot_width/2.:
			return D[1:]
		else:
			return self.X, self.Y

	def feedback(self):
		D, L = self.closest()
		if L <= self.spot_width/2.:
			draw_target(self.dimension, 
			            D[0]+self.margin, *D[1:])
		draw_target(self.dimension, self.spot_width/2., self.X, self.Y)


# rake cursor technique ######################################################

import socket

HOST = "169.254.163.247"
PORT = 1235


class RakeCursor(Flat):
	def __init__(self, distractors, distances, distance, dimension, origin,
	                   D=300, alpha=.5, **kwargs):
		self.D = D
		self.alpha = alpha
		self.Xg, self.Yg = 0, 0

		self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
		self.socket.connect((HOST, PORT))

		self.Xo, self.Yo, _ = origin
		self.Xo += X_OFFSET
		self.Yo += SCREEN_HEIGHT/2

		super(RakeCursor, self).__init__(distractors, distances, distance,
		                                 dimension, origin, **kwargs)

		self.cursor = {
			1: self.cursor_1d,
			2: self.cursor_2d,
		}[self.dimension]
		
		
	def motion(self, dx=0, dy=0, dt=1):
		self.get_gaze()
		dX, dY = self.dP(dx, dy)
		self.X += dX
		self.Y += dY
	
	def pick(self):
		self.socket.close()
		return super(RakeCursor, self).pick()

	def get_gaze(self):
		self.socket.send("a")
		data = self.socket.recv(1024)
		x, y = data.split()
		self.Xgm = self.Xg = float(x)*SCREEN_WIDTH-self.Xo
		self.Ygm = self.Yg = float(y)*SCREEN_HEIGHT-self.Yo

	def cursor_1d(self):
		d_min = float("inf")
		Xb = self.X # best cursor position
		dW = self.D * 1.
		i0 = int(floor(-(self.X+self.Xo)/dW))
		alpha = self.alpha
		for i in range(i0, i0+int(SCREEN_WIDTH/dW)+2):
			X = self.X + i*dW
			draw_cursor(self.dimension, X, 0., alpha)
			if 0 <= X+self.Xo <= SCREEN_WIDTH:
				d = abs(X - self.Xgm)
				if d < d_min:
					d_min = d
					Xb = X
		self.X = Xb
		self.Xgm = self.X
		draw_cursor(self.dimension, self.X)
	
	
	def cursor_2d(self):
		d_min = float("inf")
		Xb, Yb = self.X, self.Y # best cursor position
		dW, dH = self.D * 1., self.D * sin(pi/3.)
		i0 = int(floor(-(self.X+self.Xo)/dW))
		j0 = int(ceil(-(self.Y+self.Yo)/dH))
		alpha = self.alpha
		for i in range(i0, i0+int(SCREEN_WIDTH/dW)+2):
			for j in range(j0, j0+int(SCREEN_HEIGHT/dH)+1):
				X = self.X + i*dW + (j%2) * dW/2
				Y = self.Y + j*dH
				draw_cursor(self.dimension, X, Y, alpha)
				if (-12 <= X+self.Xo <= SCREEN_WIDTH) and (0 <= Y+self.Yo <= SCREEN_HEIGHT + 22):
					d = hypot(X - self.Xgm, Y - self.Ygm)
					if d < d_min:
						d_min = d
						Xb, Yb = X, Y
		self.X, self.Y = Xb, Yb
		self.Xgm, self.Ygm = self.X, self.Y
		draw_cursor(self.dimension, self.X, self.Y)
