//////////////////////////////////////////////////////////////////////////////
// file      : input.cpp
// author    : renaud blanch <http://iihm.imag.fr/blanch/>
// copyright : Copyright © 2010-2011 - Renaud Blanch
// licence   : GPLv3 <http://www.gnu.org/licenses/gpl.html>
// content   : device & cursor movements according to HID exposed to python
//             see <http://iihm.imag.fr/blanch/projects/distractors/>
// history   : 2005-03-03 - [rb] creation
//             2005-10-09 - [rb] refactoring
//             2006-10-01 - [rb] adapted from osx-trace/src/input.cpp
//             2009-07-28 - [rb] grabbed from sandbox/hid + code clean-up
//             2011-05-08 - [rb] prepared for distribution
//////////////////////////////////////////////////////////////////////////////


/* includes *****************************************************************/

#include <map>
#include <vector>
#include <iostream>
#include <exception>
#include <stdexcept>
using namespace std;

#include <mach/mach.h>

#include <IOKit/IOKitLib.h>
#include <IOKit/IOCFPlugIn.h>
#include <IOKit/hid/IOHIDUsageTables.h>
#include <IOKit/hid/IOHIDLib.h>

#include <Carbon/Carbon.h>


/* forward declaration ******************************************************/

static void button_cb(int);
static void motion_cb(int, int);


/* Element definition *******************************************************/

struct Element {
	int usagePage;
	int usage;
	IOHIDElementCookie cookie;
	
	Element(CFTypeRef hidElement);
};

typedef std::map< IOHIDElementCookie, Element * > ElementMap;

Element::Element(CFTypeRef hidElement) {
	CFNumberGetValue(CFNumberRef(CFDictionaryGetValue(CFDictionaryRef(hidElement), CFSTR(kIOHIDElementUsagePageKey))), kCFNumberIntType, &usagePage);
	CFNumberGetValue(CFNumberRef(CFDictionaryGetValue(CFDictionaryRef(hidElement), CFSTR(kIOHIDElementUsageKey))), kCFNumberIntType, &usage);
	CFNumberGetValue(CFNumberRef(CFDictionaryGetValue(CFDictionaryRef(hidElement), CFSTR(kIOHIDElementCookieKey))), kCFNumberIntType, &cookie);
}



/* Device definition ********************************************************/

struct Device {
	IOHIDDeviceInterface **interface;
	IOHIDQueueInterface **queue;
	ElementMap elements;

	Device(io_object_t hidDevice);
	~Device();

	void scanHIDProperties(CFMutableDictionaryRef hidProperties);
	void scanHIDElement(CFTypeRef hidElement);

	static void callback(void *target, IOReturn result, void *refcon, void *sender);
};

typedef std::vector< Device * > DeviceList;


Device::Device(io_object_t hidDevice) {
	// getting IOHIDDeviceInterface for the device
	IOCFPlugInInterface **plugInInterface = 0;
	SInt32 score = 0;
	IOReturn ioResult = IOCreatePlugInInterfaceForService(hidDevice, kIOHIDDeviceUserClientTypeID, kIOCFPlugInInterfaceID, &plugInInterface, &score);
	if(ioResult != kIOReturnSuccess) 
		throw runtime_error("Failed to create **plugInInterface via IOCreatePlugInInterfaceForService.");

	HRESULT plugInResult = (*plugInInterface)->QueryInterface(plugInInterface, CFUUIDGetUUIDBytes(kIOHIDDeviceInterfaceID), (void **)(&interface));
	if(plugInResult != S_OK) 
		throw runtime_error("Couldn't query HID class device interface from plugInInterface.");

	(*plugInInterface)->Release(plugInInterface);

	(*interface)->open(interface, 0);

	// creating event queue for the device
	queue = (*interface)->allocQueue(interface);
	if(queue == 0) 
		throw runtime_error("Unable to allocate a queue.");

	(*queue)->create(queue, 0, 8);

	// creating event loop
	CFRunLoopSourceRef loopSource = 0;

	ioResult = (*queue)->createAsyncEventSource(queue, &loopSource);
	if((ioResult != kIOReturnSuccess) || (loopSource == 0)) 
		throw runtime_error("Failed to createAsyncEventSource");

	CFRunLoopRef loop = (CFRunLoopRef)GetCFRunLoopFromEventLoop(GetMainEventLoop());
	if(!CFRunLoopContainsSource(loop, loopSource, kCFRunLoopDefaultMode)) {
		CFRunLoopAddSource(loop, loopSource, kCFRunLoopDefaultMode);
	}

	ioResult = (*queue)->setEventCallout(queue, &Device::callback, queue, (void *)(this));
	if(ioResult != kIOReturnSuccess) 
		throw runtime_error("Failed to setEventCallout");

	(*queue)->start(queue);

	// getting device properties
	CFMutableDictionaryRef hidProperties = 0;	
	if(IORegistryEntryCreateCFProperties(hidDevice, &hidProperties, kCFAllocatorDefault, kNilOptions) != KERN_SUCCESS)
		throw runtime_error("Couldn't get device properties.");

	scanHIDProperties(hidProperties);

	CFRelease(hidProperties);
}

Device::~Device() {
	// stopping the queue
	(*queue)->stop(queue);

	CFRunLoopSourceRef loopSource = (*queue)->getAsyncEventSource(queue);
	if(loopSource != 0) {
		CFRunLoopRef loop = (CFRunLoopRef)GetCFRunLoopFromEventLoop(GetMainEventLoop());
		if(CFRunLoopContainsSource(loop, loopSource, kCFRunLoopDefaultMode)) {
			CFRunLoopRemoveSource(loop, loopSource, kCFRunLoopDefaultMode);
		}
		CFRelease(loopSource);
	}

	(*queue)->dispose(queue);
	(*queue)->Release(queue);

	// closing the device
	(*interface)->close(interface);
	(*interface)->Release(interface);
}


void Device::scanHIDProperties(CFMutableDictionaryRef hidProperties) {
	CFTypeRef hidElements = CFDictionaryGetValue(hidProperties, CFSTR(kIOHIDElementKey));
	if((hidElements == 0) || (CFGetTypeID(hidElements) != CFArrayGetTypeID())) {
		cerr << "Couldn't get device elements." << endl;
		return;
	}

	int count = CFArrayGetCount(CFArrayRef(hidElements));
	for(int i = 0; i < count; i++) {
		CFTypeRef hidElement = CFArrayGetValueAtIndex(CFArrayRef(hidElements), i);
		if(CFGetTypeID(hidElement) != CFDictionaryGetTypeID()) {
			cerr << "Couldn't get element dictionnary." << endl;
			continue;
		}

		scanHIDElement(hidElement);
	}
}

void Device::scanHIDElement(CFTypeRef hidElement) {
	// getting element type
	int type = 0;
	CFNumberGetValue(CFNumberRef(CFDictionaryGetValue(CFDictionaryRef(hidElement), CFSTR(kIOHIDElementTypeKey))), kCFNumberIntType, &type);

	if(type == kIOHIDElementTypeCollection) {
		scanHIDProperties((CFMutableDictionaryRef)hidElement);

	} else {
		Element *element = new Element(hidElement);
		(*queue)->addElement(queue, element->cookie, 0);
		elements[element->cookie] = element;
	}
}


void Device::callback(void *target, IOReturn result, void *refcon, void *sender) {
	Device *self = reinterpret_cast< Device * >(refcon);
	IOHIDQueueInterface **queue = (IOHIDQueueInterface **)(target); /* == self->queue */

	static AbsoluteTime zeroTime = {0, 0};
	IOHIDEventStruct event;

	int dx = 0;
	int dy = 0;

	while((*queue)->getNextEvent(queue, &event, zeroTime, 0) == kIOReturnSuccess) {
		Element *element = self->elements[event.elementCookie];

		switch(element->usagePage) {
		case kHIDPage_Button:
			switch(element->usage) {
			case kHIDUsage_Button_1:
				button_cb(event.value != 0);
				break;
			}
			break;

		case kHIDPage_GenericDesktop:
			switch(element->usage) {
			case kHIDUsage_GD_X:
				dx += event.value;
				break;

			case kHIDUsage_GD_Y:
				dy += event.value;
				break;
			}
			break;
		}
	}
	
	if(dx != 0 || dy != 0) {
		motion_cb(dx, dy);
	}
	
	result = kIOReturnSuccess;
}


/* module implementation ****************************************************/

static mach_port_t masterPort;
static DeviceList devices;

void setupDevices(int usagePage, int usage);

void setup() {
	IOReturn ioResult = IOMasterPort(bootstrap_port, &masterPort);
	if(ioResult != kIOReturnSuccess) 
		throw runtime_error("Couldn't create a master I/O Kit Port.");

	setupDevices(kHIDPage_GenericDesktop, kHIDUsage_GD_Mouse);
	CGAssociateMouseAndMouseCursorPosition(false);
}

void cleanup() {
	for(DeviceList::iterator d = devices.begin(); d != devices.end(); ++d) {
		delete (*d);
	}

	mach_port_deallocate(mach_task_self(), masterPort);
}


void setupDevices(int usagePage, int usage) {
	CFMutableDictionaryRef hidMatchDictionary = IOServiceMatching(kIOHIDDeviceKey);
	if(hidMatchDictionary == 0)
		throw runtime_error("Failed to get HID CFMutableDictionaryRef via IOServiceMatching.");

	// Add key for device type to refine the matching dictionary.
	CFNumberRef refUsagePage = CFNumberCreate(kCFAllocatorDefault, kCFNumberIntType, &usagePage);
	CFNumberRef refUsage = CFNumberCreate(kCFAllocatorDefault, kCFNumberIntType, &usage);
	CFDictionarySetValue(hidMatchDictionary, CFSTR(kIOHIDPrimaryUsagePageKey), refUsagePage);
	CFDictionarySetValue(hidMatchDictionary, CFSTR(kIOHIDPrimaryUsageKey), refUsage);

	// Now search I/O Registry for matching devices.
	io_iterator_t hidObjectIterator;
	IOReturn ioResult = IOServiceGetMatchingServices(masterPort, hidMatchDictionary, &hidObjectIterator);
	if((ioResult != kIOReturnSuccess) || (hidObjectIterator == 0)) 
		throw runtime_error("Couldn't create a HID object iterator.");

	io_object_t hidDevice = 0;
	while((hidDevice = IOIteratorNext(hidObjectIterator))) {
		Device *device = new Device(hidDevice);
		if(device != 0) {
			devices.push_back(device);
		}
	}

	IOObjectRelease(hidObjectIterator);
}


/* python binding ***********************************************************/

#include <Python.h>

// python has no finalizer for modules
struct Finalizer { ~Finalizer() { cleanup(); } };
static Finalizer finalizer;


// python callbacks
static PyObject *_input_motion_cb = NULL; //Py_None;
static PyObject *_input_button_cb = NULL; //Py_None;

static PyObject *_input_set_motion_cb(PyObject *self, PyObject *args) {
	PyObject *temp;
	
	if(PyArg_ParseTuple(args, "O:set_motion_cb", &temp)) {
		if(!PyCallable_Check(temp) && temp != Py_None) {
			PyErr_SetString(PyExc_TypeError, "parameter must be callable or None");
			return NULL;
		}
		Py_XINCREF(temp);
		Py_XDECREF(_input_motion_cb);
		_input_motion_cb = temp;
		
		Py_RETURN_NONE;
	}
	
	return NULL;
}

static PyObject *_input_set_button_cb(PyObject *self, PyObject *args) {
	PyObject *temp;
	
	if(PyArg_ParseTuple(args, "O:set_button_cb", &temp)) {
		if(!PyCallable_Check(temp) && temp != Py_None) {
			PyErr_SetString(PyExc_TypeError, "parameter must be callable or None");
			return NULL;
		}
		Py_XINCREF(temp);
		Py_XDECREF(_input_button_cb);
		_input_button_cb = temp;
		
		Py_RETURN_NONE;
	}
	
	return NULL;
}


// C callbacks
static void button_cb(int down) {
	static PyGILState_STATE gstate;
	gstate = PyGILState_Ensure();

	if(_input_button_cb != Py_None) {
		PyObject *arglist = Py_BuildValue("(i)", down);
		PyObject *result = PyEval_CallObject(_input_button_cb, arglist);
		Py_DECREF(arglist);
		if(result != NULL) {
			Py_DECREF(result);
		}
	}

	PyGILState_Release(gstate);
}

static void motion_cb(int dx, int dy) {
	static PyGILState_STATE gstate;
	gstate = PyGILState_Ensure();

	if(_input_motion_cb != Py_None) {
		PyObject *arglist = Py_BuildValue("(i,i)", dx, dy);
		PyObject *result = PyEval_CallObject(_input_motion_cb, arglist);
		Py_DECREF(arglist);
		if(result != NULL) {
			Py_DECREF(result);
		}
	}

	PyGILState_Release(gstate);
}


// module initialization
static PyMethodDef InputMethods[] = {
	{"set_motion_cb", _input_set_motion_cb, METH_VARARGS, "Set physical device motion callback."},
	{"set_button_cb", _input_set_button_cb, METH_VARARGS, "Set physical device button callback."},
	{NULL, NULL, 0, NULL}
};

PyMODINIT_FUNC init_input(void) {
	_input_motion_cb = Py_None;
	Py_XINCREF(_input_motion_cb);
	_input_button_cb = Py_None;
	Py_XINCREF(_input_button_cb);

	(void)Py_InitModule("_input", InputMethods);
	setup();
}

